@@ -14,19 +14,20 @@
 --    /Backtracking, Interleaving, and Terminating
 --        Monad Transformers/, by
 --    Oleg Kiselyov, Chung-chieh Shan, Daniel P. Friedman, Amr Sabry
---    (<http://www.cs.rutgers.edu/~ccshan/logicprog/LogicT-icfp2005.pdf>)
+--    (<http://okmij.org/ftp/papers/LogicT.pdf>)
 -------------------------------------------------------------------------
 
-module Control.Monad.Logic.Class (MonadLogic(..), reflect, lnot) where
+{-# LANGUAGE CPP  #-}
 
-import qualified Control.Monad.State.Lazy as LazyST
-import qualified Control.Monad.State.Strict as StrictST
+#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE Safe #-}
+#endif
+
+module Control.Monad.Logic.Class (MonadLogic(..), reflect) where
 
 import Control.Monad.Reader
-
-import Data.Monoid
-import qualified Control.Monad.Writer.Lazy as LazyWT
-import qualified Control.Monad.Writer.Strict as StrictWT
+import qualified Control.Monad.State.Lazy as LazyST
+import qualified Control.Monad.State.Strict as StrictST
 
 -------------------------------------------------------------------------------
 -- | Minimal implementation: msplit
@@ -79,6 +80,10 @@ class (MonadPlus m) => MonadLogic m where
     --   such.
     once       :: m a -> m a
 
+    -- | Inverts a logic computation. If @m@ succeeds with at least one value,
+    -- @lnot m@ fails. If @m@ fails, then @lnot m@ succeeds the value @()@.
+    lnot :: m a -> m ()
+
     -- All the class functions besides msplit can be derived from msplit, if
     -- desired
     interleave m1 m2 = msplit m1 >>=
@@ -92,6 +97,8 @@ class (MonadPlus m) => MonadLogic m where
     once m = do (a, _) <- maybe mzero return =<< msplit m
                 return a
 
+    lnot m = ifte (once m) (const mzero) (return ())
+
 
 -------------------------------------------------------------------------------
 -- | The inverse of msplit. Satisfies the following law:
@@ -101,33 +108,27 @@ reflect :: MonadLogic m => Maybe (a, m a) -> m a
 reflect Nothing = mzero
 reflect (Just (a, m)) = return a `mplus` m
 
--- | Inverts a logic computation. If @m@ succeeds with at least one value,
--- @lnot m@ fails. If @m@ fails, then @lnot m@ succeeds the value @()@.
-lnot :: MonadLogic m => m a -> m ()
-lnot m = ifte (once m) (const mzero) (return ())
-
 -- An instance of MonadLogic for lists
 instance MonadLogic [] where
     msplit []     = return Nothing
     msplit (x:xs) = return $ Just (x, xs)
 
--- Some of these may be questionable instances. Splitting a transformer does
--- not allow you to provide different input to the monadic object returned.
--- So, for instance, in:
+-- | Note that splitting a transformer does
+-- not allow you to provide different input
+-- to the monadic object returned.
+-- For instance, in:
 --
---  let Just (_, rm') = runReaderT (msplit rm) r
---   in runReaderT rm' r'
+-- > let Just (_, rm') = runReaderT (msplit rm) r in runReaderT rm' r'
 --
--- The "r'" parameter will be ignored, as "r" was already threaded through the
--- computation. The results are similar for StateT. However, this is likely not
--- an issue as most uses of msplit (all the ones in this library, at least) would
--- not allow for that anyway.
+-- @r'@ will be ignored, because @r@ was already threaded through the
+-- computation.
 instance MonadLogic m => MonadLogic (ReaderT e m) where
     msplit rm = ReaderT $ \e -> do r <- msplit $ runReaderT rm e
                                    case r of
                                      Nothing -> return Nothing
                                      Just (a, m) -> return (Just (a, lift m))
 
+-- | See note on splitting above.
 instance MonadLogic m => MonadLogic (StrictST.StateT s m) where
     msplit sm = StrictST.StateT $ \s ->
                     do r <- msplit (StrictST.runStateT sm s)
@@ -148,6 +149,7 @@ instance MonadLogic m => MonadLogic (StrictST.StateT s m) where
 
     once ma = StrictST.StateT $ \s -> once (StrictST.runStateT ma s)
 
+-- | See note on splitting above.
 instance MonadLogic m => MonadLogic (LazyST.StateT s m) where
     msplit sm = LazyST.StateT $ \s ->
                     do r <- msplit (LazyST.runStateT sm s)
@@ -167,47 +169,3 @@ instance MonadLogic m => MonadLogic (LazyST.StateT s m) where
                                               (LazyST.runStateT el s)
 
     once ma = LazyST.StateT $ \s -> once (LazyST.runStateT ma s)
-
-instance (MonadLogic m, Monoid w) => MonadLogic (StrictWT.WriterT w m) where
-    msplit wm = StrictWT.WriterT $
-                    do r <- msplit (StrictWT.runWriterT wm)
-                       case r of
-                            Nothing -> return (Nothing, mempty)
-                            Just ((a,w), m) ->
-                                return (Just (a, StrictWT.WriterT m), w)
-
-    interleave ma mb = StrictWT.WriterT $
-                        StrictWT.runWriterT ma `interleave` StrictWT.runWriterT mb
-
-    ma >>- f = StrictWT.WriterT $
-                StrictWT.runWriterT ma >>- \(a,w) ->
-                    StrictWT.runWriterT (StrictWT.tell w >> f a)
-
-    ifte t th el = StrictWT.WriterT $
-                    ifte (StrictWT.runWriterT t)
-                         (\(a,w) -> StrictWT.runWriterT (StrictWT.tell w >> th a))
-                         (StrictWT.runWriterT el)
-
-    once ma = StrictWT.WriterT $ once (StrictWT.runWriterT ma)
-
-instance (MonadLogic m, Monoid w) => MonadLogic (LazyWT.WriterT w m) where
-    msplit wm = LazyWT.WriterT $
-                    do r <- msplit (LazyWT.runWriterT wm)
-                       case r of
-                            Nothing -> return (Nothing, mempty)
-                            Just ((a,w), m) ->
-                                return (Just (a, LazyWT.WriterT m), w)
-
-    interleave ma mb = LazyWT.WriterT $
-                        LazyWT.runWriterT ma `interleave` LazyWT.runWriterT mb
-
-    ma >>- f = LazyWT.WriterT $
-                LazyWT.runWriterT ma >>- \(a,w) ->
-                    LazyWT.runWriterT (LazyWT.tell w >> f a)
-
-    ifte t th el = LazyWT.WriterT $
-                    ifte (LazyWT.runWriterT t)
-                         (\(a,w) -> LazyWT.runWriterT (LazyWT.tell w >> th a))
-                         (LazyWT.runWriterT el)
-
-    once ma = LazyWT.WriterT $ once (LazyWT.runWriterT ma)
